/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**

* \file bt_l2cap_cl-cmds.h
* 
* \author Kevin Martin <kevmarti@tik.ee.ethz.ch>
* 
* \author Martin Hinz <btnode@hinz.ch>

* \brief Provides terminal commands for bt_l2cap_cl communication.
* 
* cmds for sending text to other connection, getting packets and to change
* the pkt type to avoid fragmentation!\n
* 
* Commands:\n
 - send: To send some bytes over a connection with choosen service nr\n
 - get: To get a pkt with a given service nr and waiting timeout\n
 - chpkttype: To change pkt type to #BT_HCI_PACKET_TYPE_DM3
* 
* @see \link btn-terminal.h btn-terminal.h \endlink
*/

#include <bt/l2cap_cl.h>

/**
 * \brief Initializes the commands.
 */
void l2cap_cl_cmds_init(struct btstack* bt_stack);

/*!
 * @brief Registers the commands into the terminal. 
 * 
 * Call acl_com_cmds_init() first
 * in order to use them.
 */
void l2cap_cl_cmds_register_cmds(void);
